/**
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * beck et al. projects GmbH
 * Theresienhoehe 13
 * 80336 Munich
 * Phone: +49-89-5442530
 * mailto: info@bea.de
 */
package de.bsvrz.ars.export.prot;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.Stack;

import de.bsvrz.ars.export.ExportProfile;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.DoubleAttributeType;
import de.bsvrz.dav.daf.main.config.IntegerAttributeType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pat.onlprot.standardProtocolModule.ClientProtocollerInterface;
import de.bsvrz.pat.onlprot.standardProtocolModule.ProtocolModule;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Protokollierungsmodul für Datenbankexporte
 * 
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @version $Revision: 1.1 $ / $Date: 2008/01/22 16:39:40 $ / ($Author: yvonnes $)
 */
public class DatabaseProtocoller extends ProtocolModule {
    private static final DateFormat dateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");
    private SqlProperties sqlProperties;


    /**
     * 
     */
    public DatabaseProtocoller() {
        sqlProperties = ExportProfile.SINGLETON_INSTANCE.getSqlProperties();        
        setProtocoller(new SqlProtocoller());
    }


    /* (non-Javadoc)
     * @see ProtocolModuleInterface#initProtocol(ArgumentList, java.io.PrintWriter, java.lang.String[])
     */
    public ClientReceiverInterface initProtocol(ArgumentList argumentList, PrintWriter protocolFile, String[] args) {
        setProtocolFileWriter(protocolFile);
        getProtocoller().writeHeader(args);
        return getProtocoller();
    }


    /* (non-Javadoc)
     * @see ProtocolModuleInterface#closeProtocol()
     */
    public void closeProtocol() {
        getProtocoller().writeFooter();
        getProtocolFileWriter().flush();
        getProtocolFileWriter().close();
    }


    /* (non-Javadoc)
     * @see ProtocolModuleInterface#getHelp()
     */
    public String getHelp() {
        return "help";
    }


    /* (non-Javadoc)
     * @see ProtocolModule#update(ResultData[])
     */
    public void update(ResultData[] results) {
    }


    // Inner Class
    public class SqlProtocoller implements ClientProtocollerInterface {
        /** SQL Kommentarzeichen + ein Leerzeichen */
        public static final String SQL_CMMNT = "-- ";

        /** SQL Trennlinie (ohne newline) */
        public static final String SQL_SEPARATOR = "-- ---------------------------------------------------------";

        /**
         * Alle SQL-Statements als printf-Strings mit ";" und "newline" am Ende
         */
        private final String INSERT_INTO_EXPORTINFO = "INSERT INTO exportinfo (name, wert)   VALUES (%s, %s);\n";
        private final String INSERT_INTO_ATTRIBUT = "  INSERT INTO attribut (attribut_pk, name, wertZahl, wertKommazahl, wertBool, wertText, archivdaten_fk, feld_fk, liste_fk)   "
                + "VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s);\n";
        private final String INSERT_INTO_FELD = "  INSERT INTO feld (feld_pk, name, laenge, liste_fk, archivdaten_fk)    VALUES (%s, %s, %s, %s, %s);\n";
        private final String INSERT_INTO_LISTE = "  INSERT INTO liste (liste_pk, name, liste_fk, feld_fk, archivdaten_fk)    VALUES (%s, %s, %s, %s, %s);\n";
        /**
         * 
         */
        private final String INSERT_INTO_ARCHIVDATEN = "INSERT INTO archivdaten (archivdaten_pk, objekt_id, objekt_pid, attributgruppe_id, attributgruppe_pid, aspekt_id, aspekt_pid, simulationsvariante, datenzeit)   "
                + "VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s);\n";

        /** Type und Stack über die wir uns merken, in welcher Verschachtelung wir gerade sind */
        private static final int ARCHIVDATEN = 0;
        private static final int FELD = 1;
        private static final int LISTE = 2;
        private static final int ATTRIBUT = 3;
        private Stack<Integer> stack;

        /** Ints, die die Unique IDs für die Primary/Foreign-Keys darstellen */
        private int archivdatenPK, feldPK, listePK, attributPK;


        /**
         * 
         */
        public SqlProtocoller() {
            archivdatenPK = 0;
            feldPK = 0;
            listePK = 0;
            attributPK = 0;
            stack = new Stack<Integer>();
        }


        /* (non-Javadoc)
         * @see ClientProtocollerInterface#writeHeader(java.lang.String[])
         */
        public void writeHeader(String[] args) {
            getProtocolFileWriter().println(SQL_CMMNT + "VRZ3/DatenExport vom " + dateFormat.format(new Date(System.currentTimeMillis())));
            getProtocolFileWriter().println(SQL_SEPARATOR);
            for (String arg : args)
                getProtocolFileWriter().println(SQL_CMMNT + arg);
            getProtocolFileWriter().println(SQL_SEPARATOR);

            // XXX Error checks etc
            try {
                BufferedReader br = new BufferedReader(new InputStreamReader(this.getClass().getResourceAsStream("/de/bsvrz/ars/export/resources/mysql_ddl.sql")));
                StringBuffer sb = new StringBuffer(2048);
                String s;
                while ((s = br.readLine()) != null)
                    sb.append(s + Debug.NEWLINE);
                getProtocolFileWriter().println(sqlProperties.parseDdl(sb.toString()));
            } catch (IOException e) {
            }
        }


        /* (non-Javadoc)
         * @see ClientProtocollerInterface#writeFooter()
         */
        public void writeFooter() {
            getProtocolFileWriter().println(SQL_SEPARATOR);
            getProtocolFileWriter().println(SQL_CMMNT + "Ende der Datei");
        }


        /* (non-Javadoc)
         * @see ClientReceiverInterface#update(ResultData[])
         */
        public void update(ResultData[] results) {
            stack.push(ARCHIVDATEN);
            for (int i = 0; i < results.length; i++) {
                handleArchiveData(results[i]);
                if (results[i].hasData()) {
                    Iterator itr = results[i].getData().iterator();
                    while (itr.hasNext())
                        handlePayload((Data) itr.next());
                }
            }
            stack.pop();
        }


        private void handlePayload(Data data) {
            if (data != null) {
                if (data.isPlain()) {
                    stack.push(ATTRIBUT);
                    handleAttribute(data.getName(), data.valueToString(), data.getAttributeType());
                    stack.pop();
                } else if (data.isArray()) {
                    stack.push(FELD);
                    handleArray(data.getName(), data.asArray().getLength());
                    Iterator i = data.iterator();
                    while (i.hasNext())
                        handlePayload((Data) i.next());
                    stack.pop();
                } else if (data.isList()) {
                    stack.push(LISTE);
                    handleList(data.getName());
                    Iterator i = data.iterator();
                    while (i.hasNext())
                        handlePayload((Data) i.next());
                    stack.pop();
                }
            }
        }


        private void handleArchiveData(ResultData result) {
            archivdatenPK++;
            SystemObject o = result.getObject();
            DataDescription dd = result.getDataDescription();

            getProtocolFileWriter().printf(INSERT_INTO_ARCHIVDATEN, archivdatenPK, genValueNumberOrNull(o.getId()), genValueQuotedOrNull(o.getPid()),
                    genValueNumberOrNull(dd.getAttributeGroup().getId()), genValueQuotedOrNull(dd.getAttributeGroup().getPid()),
                    genValueNumberOrNull(dd.getAspect().getId()), genValueQuotedOrNull(dd.getAspect().getPid()), dd.getSimulationVariant(),
                    genValueNumberOrNull(result.getDataTime()));
        }


        private void handleArray(String name, int length) {
            feldPK++;
            // Parent ermitteln
            int liste = 0, archivdaten = 0;
            if (stack.get(stack.size() - 2) == LISTE)
                liste = listePK;
            else if (stack.get(stack.size() - 2) == ARCHIVDATEN)
                archivdaten = archivdatenPK;
            else
                System.out.println("ERROR: Falscher Parent für " + name);
            getProtocolFileWriter().printf(INSERT_INTO_FELD, feldPK, genValueQuotedOrNull(name), length, genValueNumberOrNull(liste),
                    genValueNumberOrNull(archivdaten));
        }


        private void handleList(String name) {
            listePK++;
            // Parent ermitteln
            int liste = 0, feld = 0, archivdaten = 0;
            if (stack.get(stack.size() - 2) == LISTE)
                liste = listePK;
            else if (stack.get(stack.size() - 2) == FELD)
                feld = feldPK;
            else if (stack.get(stack.size() - 2) == ARCHIVDATEN)
                archivdaten = archivdatenPK;
            else 
                System.out.println("ERROR: Falscher Parent für " + name);
            getProtocolFileWriter().printf(INSERT_INTO_LISTE, listePK, genValueQuotedOrNull(name), genValueNumberOrNull(liste), genValueNumberOrNull(feld),
                    genValueNumberOrNull(archivdaten));
        }


        private void handleAttribute(String name, String wert, AttributeType type) {
            attributPK++;
            // Parent ermitteln
            int liste = 0, feld = 0, archivdaten = 0;
            if (stack.get(stack.size() - 2) == LISTE) 
                liste = listePK;
            else if (stack.get(stack.size() - 2) == FELD) 
                feld = feldPK;
            else if (stack.get(stack.size() - 2) == ARCHIVDATEN) 
                archivdaten = archivdatenPK;
            else 
                System.out.println("ERROR: Falscher Parent für " + name);
            

            // Typ ermitteln
            // Wir müssen sicher gehen, dass sich die Werte als entsprechender Typ parsen
            // lassen -> dann aber wieder Text daraus machen für die SQL Insert Statements
            boolean isInt = false, isDouble = false, isBool = false, isText = false;
            String intVal, doubleVal, boolVal;
            if (type instanceof IntegerAttributeType) {
                if (type.getName().equalsIgnoreCase("janein")) {
                    // Boolsche Werte sind 0 und 1
                    try {
                        if (Integer.parseInt(wert) == 1)
                            boolVal = sqlProperties.getBooleanTrue();
                        else
                            boolVal = sqlProperties.getBooleanFalse();
                        isBool = true;
                    } catch (Exception e) {
                        isText = true;
                    }
                } else {
                    // Normaler Int-Wert
                    try {
                        intVal = "" + Integer.parseInt(wert);
                        isInt = true;
                    } catch (Exception e) {
                        isText = true;
                    }
                }
            } else if (type instanceof DoubleAttributeType) {
                try {
                    doubleVal = "" + Double.parseDouble(wert);
                    isDouble = true;
                } catch (Exception e) {
                    isText = true;
                }
            }

            // sicherstellen, dass nur einer gesetzt ist, andernfalls text
            if ((isInt && (isDouble || isBool)) || (isDouble && isBool))
                isText = true;

            // wenn nichts gesetzt ist, ebenfalls in textform ausgeben
            if (!isInt && !isDouble && !isBool)
                isText = true;

            if (isText)
                getProtocolFileWriter().printf(INSERT_INTO_ATTRIBUT, attributPK, genValueQuotedOrNull(name), "null", "null", "null",
                        genValueQuotedOrNull(wert), genValueNumberOrNull(archivdaten), genValueNumberOrNull(feld), genValueNumberOrNull(liste));
            else if (isInt)
                getProtocolFileWriter().printf(INSERT_INTO_ATTRIBUT, attributPK, genValueQuotedOrNull(name), genValueQuotedOrNull(wert), "null", "null",
                        "null", genValueNumberOrNull(archivdaten), genValueNumberOrNull(feld), genValueNumberOrNull(liste));
            else if (isBool)
                getProtocolFileWriter().printf(INSERT_INTO_ATTRIBUT, attributPK, genValueQuotedOrNull(name), "null", "null", genValueQuotedOrNull(wert),
                        "null", genValueNumberOrNull(archivdaten), genValueNumberOrNull(feld), genValueNumberOrNull(liste));
            else if (isDouble)
                getProtocolFileWriter().printf(INSERT_INTO_ATTRIBUT, attributPK, genValueQuotedOrNull(name), "null", genValueQuotedOrNull(wert), "null",
                        genValueQuotedOrNull(wert), genValueNumberOrNull(archivdaten), genValueNumberOrNull(feld), genValueNumberOrNull(liste));
            else
                System.out.println("ERROR!!!"); // dürfte nie vorkommen wegen checks oben
        }


        private String genValueQuotedOrNull(String input) {
            return (input == null) ? "null" : "'" + input + "'";
        }


        private String genValueNumberOrNull(long input) {
            return (input == 0) ? "null" : "" + input;
        }
    }
}
